@@ -15,15 +15,15 @@
 //! see the [`log`]'s package [`LevelFilter`] type for available values.
 //!
 //! ```bash
-//! # In your shell of choose:
+//! ## In your shell of your choice:
 //!
-//! # Set the log severity to only print log message with info severity or
-//! # higher, trace and debug messages won't be printed anymore.
+//! ## Set the log severity to only print log messages with info severity or
+//! ## higher, trace and debug messages won't be printed anymore.
 //! $ LOG=info ./my_binary
 //!
-//! # Set the log severity to only print log message with warning severity or
-//! # higher, informational (or lower severity) messages won't be printed
-//! # anymore.
+//! ## Set the log severity to only print log messages with warning severity or
+//! ## higher, informational (or lower severity) messages won't be printed
+//! ## anymore.
 //! $ LOG=warn ./my_binary
 //! ```
 //!
@@ -32,18 +32,23 @@
 //! set the severity to debug.
 //!
 //! ```bash
-//! # In your shell of choose:
+//! ## In your shell of your choice:
 //!
-//! # Enables trace logging.
+//! ## Enables trace logging.
 //! $ TRACE=1 ./my_binary
 //!
-//! # Enables debug logging.
+//! ## Enables debug logging.
 //! $ DEBUG=1 ./my_binary
 //! ```
 //!
 //! If none of these environment variables are found it will default to an
 //! information severity.
 //!
+//! In addition to these runtime filters the [log] crate provides [compile time
+//! filters] which allows you to filter log messages at compile time.
+//!
+//! [log]: mod@log
+//! [compile time filters]: mod@log#compile-time-filters
 //!
 //! # Logging requests
 //!
@@ -69,20 +74,20 @@
 //! `LOG_TARGET` environment variable.
 //!
 //! ```bash
-//! # In your shell of choose:
+//! ## In your shell of choose:
 //!
-//! # Only log messages from your crate.
+//! ## Only log messages from your crate.
 //! $ LOG_TARGET=my_crate ./my_binary
 //!
-//! # Only log messages from the `my_module` module in your crate.
+//! ## Only log messages from the `my_module` module in your crate.
 //! $ LOG_TARGET=my_crate::my_module ./my_binary
 //!
-//! # Multiple log targets are also supported by separating the values by a
-//! # comma.
+//! ## Multiple log targets are also supported by separating the values by a
+//! ## comma.
 //! $ LOG_TARGET=my_crate::my_module,my_crate::my_other_module ./my_binary
 //!
-//! # Very useful in combination with trace severity to get all messages you
-//! # want, but filter out the message you don't need.
+//! ## Very useful in combination with trace severity to get all messages you
+//! ## want, but filter out the message you don't need.
 //! $ LOG_LEVEL=trace LOG_TARGET=my_crate::my_module ./my_binary
 //! ```
 //!
@@ -93,8 +98,12 @@
 //!
 //! # Format
 //!
-//! The format follows the [logfmt] format. For regular messages, printed to
-//! standard error, the following format is used:
+//! This crate supports two formats [logfmt] and [gcloud]. In the examples below
+//! we'll use the logfmt format as the author believes its easier to read for
+//! humans.
+//!
+//! For regular messages, printed to standard error, the following format is
+//! used:
 //!
 //! ```text
 //! ts="YYYY-MM-DDTHH:MM:SS.MICROSZ" lvl="$LOG_LEVEL" msg="$message" target="$target" module="$module"
@@ -119,6 +128,7 @@
 //! enabled, this feature is enabled by default, see [Timestamp feature] below.
 //!
 //! [logfmt]: https://www.brandur.org/logfmt
+//! [gcloud]: https://cloud.google.com/logging/docs/structured-logging
 //!
 //!
 //! # Crate features
@@ -232,13 +242,19 @@
 #![cfg_attr(all(feature = "log-panic", feature = "nightly"), feature(backtrace))]
 
 use std::cell::RefCell;
-use std::env;
 use std::io::{self, IoSlice, Write};
+use std::marker::PhantomData;
 
-use log::{LevelFilter, Log, Metadata, Record, SetLoggerError};
+use log::{kv, LevelFilter, Log, Metadata, Record, SetLoggerError};
 
 mod format;
-use format::{Buffer, BUFS_SIZE};
+use format::{Buffer, Format, BUFS_SIZE};
+
+mod config;
+pub use config::Config;
+
+#[cfg(feature = "timestamp")]
+mod timestamp;
 
 #[cfg(test)]
 mod tests;
@@ -270,114 +286,29 @@ macro_rules! request {
 #[doc(hidden)]
 pub use log as _log;
 
-/// Initialise the logger.
+/// Convenience method to quick initialise the logger.
 ///
-/// See the [crate level documentation] for more.
-///
-/// [crate level documentation]: index.html
-///
-/// # Panics
-///
-/// This will panic if the logger fails to initialise. Use [`try_init`] if you
-/// want to handle the error yourself.
+/// This is shortcut for `Config::logfmt().init()`.
 pub fn init() {
-    try_init().unwrap_or_else(|err| panic!("failed to initialise the logger: {}", err));
+    Config::logfmt().init();
 }
 
-/// Try to initialise the logger.
+/// Convenience method to quick initialise the logger.
 ///
-/// Unlike [`init`] this doesn't panic when the logger fails to initialise. See
-/// the [crate level documentation] for more.
-///
-/// [`init`]: fn.init.html
-/// [crate level documentation]: index.html
+/// This is shortcut for `Config::logfmt().try_init()`.
 pub fn try_init() -> Result<(), SetLoggerError> {
-    let filter = get_max_level();
-    let targets = get_log_targets();
-    let logger = Logger { filter, targets };
-    log::set_boxed_logger(Box::new(logger))?;
-    log::set_max_level(filter);
-
-    #[cfg(all(feature = "log-panic", not(feature = "nightly")))]
-    log_panics::init();
-    #[cfg(all(feature = "log-panic", feature = "nightly"))]
-    std::panic::set_hook(Box::new(log_panic));
-    Ok(())
-}
-
-/// Get the maximum log level based on the environment.
-fn get_max_level() -> LevelFilter {
-    for var in &["LOG", "LOG_LEVEL"] {
-        if let Ok(level) = env::var(var) {
-            if let Ok(level) = level.parse() {
-                return level;
-            }
-        }
-    }
-
-    if env::var("TRACE").is_ok() {
-        LevelFilter::Trace
-    } else if env::var("DEBUG").is_ok() {
-        LevelFilter::Debug
-    } else {
-        LevelFilter::Info
-    }
-}
-
-/// Get the targets to log, if any.
-fn get_log_targets() -> Targets {
-    match env::var("LOG_TARGET") {
-        Ok(ref targets) if !targets.is_empty() => {
-            Targets::Only(targets.split(',').map(|target| target.into()).collect())
-        }
-        _ => Targets::All,
-    }
-}
-
-/// Panic hook that logs the panic using [`log::error!`].
-#[cfg(all(feature = "log-panic", feature = "nightly"))]
-fn log_panic(info: &std::panic::PanicInfo<'_>) {
-    use std::backtrace::Backtrace;
-    use std::thread;
-
-    let thread = thread::current();
-    let thread_name = thread.name().unwrap_or("unnamed");
-    let msg = match info.payload().downcast_ref::<&'static str>() {
-        Some(s) => *s,
-        None => match info.payload().downcast_ref::<String>() {
-            Some(s) => &**s,
-            None => "<unknown>",
-        },
-    };
-    let (file, line) = match info.location() {
-        Some(location) => (location.file(), location.line()),
-        None => ("<unknown>", 0),
-    };
-    let backtrace = Backtrace::force_capture();
-
-    log::logger().log(
-        &Record::builder()
-            .args(format_args!(
-                // NOTE: we include file in here because it's only logged when
-                // debug severity is enabled.
-                "thread '{}' panicked at '{}', {}:{}",
-                thread_name, msg, file, line
-            ))
-            .level(log::Level::Error)
-            .target("panic")
-            .file(Some(file))
-            .line(Some(line))
-            .key_values(&("backtrace", &backtrace as &dyn std::fmt::Display))
-            .build(),
-    );
+    Config::logfmt().try_init()
 }
 
 /// Our `Log` implementation.
-struct Logger {
+struct Logger<F, Kvs> {
     /// The filter used to determine what messages to log.
     filter: LevelFilter,
     /// What logging targets to log.
     targets: Targets,
+    /// Key-values supplied for all logs.
+    kvs: Kvs,
+    _format: PhantomData<F>,
 }
 
 #[derive(Debug, Eq, PartialEq)]
@@ -408,14 +339,18 @@ impl Targets {
     }
 }
 
-impl Log for Logger {
+impl<F, Kvs> Log for Logger<F, Kvs>
+where
+    F: Format + Sync + Send,
+    Kvs: kv::Source + Sync + Send,
+{
     fn enabled(&self, metadata: &Metadata) -> bool {
         self.filter >= metadata.level() && self.targets.should_log(metadata.target())
     }
 
     fn log(&self, record: &Record) {
         if self.enabled(record.metadata()) {
-            log(record, self.filter >= LevelFilter::Debug);
+            log::<F, Kvs>(record, &self.kvs, self.filter >= LevelFilter::Debug);
         }
     }
 
@@ -425,7 +360,7 @@ impl Log for Logger {
 }
 
 /// The actual logging of a record.
-fn log(record: &Record, debug: bool) {
+fn log<F: Format, Kvs: kv::Source>(record: &Record, kvs: &Kvs, debug: bool) {
     // Thread local buffer for logging. This way we only lock standard out/error
     // for a single writev call and don't create half written logs.
     thread_local! {
@@ -437,7 +372,7 @@ fn log(record: &Record, debug: bool) {
         match buf.try_borrow_mut() {
             Ok(mut buf) => {
                 // NOTE: keep in sync with the `Err` branch below.
-                let bufs = format::record(&mut bufs, &mut buf, record, debug);
+                let bufs = F::format(&mut bufs, &mut buf, record, kvs, debug);
                 match record.target() {
                     REQUEST_TARGET => write_once(stdout(), bufs),
                     _ => write_once(stderr(), bufs),
@@ -446,13 +381,13 @@ fn log(record: &Record, debug: bool) {
             }
             Err(_) => {
                 // NOTE: We only get to this branch if we're panicking while
-                // calling `format::record`, e.g. when a `fmt::Display` impl in
-                // the `record` panics, and the `log-panic` feature is enabled
-                // which calls `error!` and in turn this function again, while
-                // still borrowing `BUF`.
+                // calling `F::format`, e.g. when a `fmt::Display` impl in the
+                // `record` panics, and the `log-panic` feature is enabled which
+                // calls `error!` and in turn this function again, while still
+                // borrowing `BUF`.
                 let mut buf = Buffer::new();
                 // NOTE: keep in sync with the `Ok` branch above.
-                let bufs = format::record(&mut bufs, &mut buf, record, debug);
+                let bufs = F::format(&mut bufs, &mut buf, record, kvs, debug);
                 match record.target() {
                     REQUEST_TARGET => write_once(stdout(), bufs),
                     _ => write_once(stderr(), bufs),
